@@ -5,6 +5,11 @@ import logging
 from copy import deepcopy
 from datetime import datetime, date
 
+try:
+    from collections import OrderedDict
+except ImportError:
+    from ordereddict import OrderedDict
+
 from .tree import Node
 
 
@@ -25,6 +30,22 @@ class SafeString(str):
 class SafeUnicode(unicode):
     pass
 
+def process_special_words(value, words=None):
+    words = words or SPECIAL_WORDS
+    for w in words:
+        value = re.sub(r'(\A|\s+)(%s)(\s+|\Z)' % w, lambda m: m.group(0).lower(), value)
+    return value
+
+def process_special_characters(value, chars=None):
+    chars = chars or SPECIAL_CHARACTERS
+    for c in chars:
+        value = value.replace(c, r'\%s' % c)
+    return value
+
+def contains_special_characters(value, chars=None):
+    chars = chars or SPECIAL_CHARACTERS
+    return any(map(lambda c: c in value, chars))
+
 def safe_solr_input(value):
     if isinstance(value, (SafeString, SafeUnicode)):
         return value
@@ -32,11 +53,9 @@ def safe_solr_input(value):
     if not isinstance(value, basestring):
         value = unicode(value)
     
-    for w in SPECIAL_WORDS:
-        value = re.sub(r'(\A|\s+)(%s)(\s+|\Z)' % w, lambda m: m.group(0).lower(), value)
+    value = process_special_words(value)
 
-    for c in SPECIAL_CHARACTERS:
-        value = value.replace(c, r'\%s' % c)
+    value = process_special_characters(value)
 
     if isinstance(value, unicode):
         return SafeUnicode(value)
@@ -90,61 +109,88 @@ class X(Node):
         obj.negate()
         return obj
 
-class LocalParams(object):
-    def __init__(self, params=None):
-        self.local_params = []
-        self.keys = set()
-
-        params = params or []
-        if isinstance(params, dict):
-            params = sorted(params.items(), key=lambda p: p[0])
-            
-        for p in params:
-            self.add(p)
-
-    def add(self, key, value=None):
-        if value is None and isinstance(key, (list, tuple)):
-            key, value = key
-        self.local_params.append((key, value))
-        self.keys.add(key)
-
-    def get(self, key, default=None):
-        for k, v in self.local_params:
-            if k == key:
-                return v
-        return default
-
-    def update(self, local_params):
-        for k, v in local_params:
-            self.add(k, v)
+class LocalParams(OrderedDict):
+    SPECIAL_CHARACTERS = " '" + SPECIAL_CHARACTERS
+    
+    def update(self, other=None, **kwargs):
+        if other is None:
+            other = OrderedDict()
+        elif isinstance(other, LocalParams):
+            other = OrderedDict(other)
+        elif hasattr(other, 'keys'):
+            other = OrderedDict((k, other[k]) for k in sorted(other.keys()))
+        elif hasattr(other, '__iter__'):
+            _other = []
+            for p in other:
+                if isinstance(p, (list, tuple)):
+                    _other.append(p)
+                else:
+                    _other.append(('type', p))
+            other = OrderedDict(_other)
+        else:
+            other = OrderedDict(type=other)
         
-    def __contains__(self, key):
-        return key in self.keys
+        other.update(sorted(kwargs.items(), key=lambda p: p[0]))
+        for k, v in other.items():
+            self.add(k, v)
+    # replace OrderedDict.__update to fix lp.update(['dismax'])
+    _OrderedDict__update = update
+    
+    def add(self, key, value=None):
+        if value is None:
+            if isinstance(key, (list, tuple)):
+                key, value = key
+            else:
+                key, value = 'type', key
+        
+        if contains_special_characters(key, self.SPECIAL_CHARACTERS):
+            raise ValueError("Key '%s' contents special characters" % key)
+        if key == 'type' and contains_special_characters(value, self.SPECIAL_CHARACTERS):
+            raise ValueError("Type value '%s' contents special characters" % value)
+        
+        self[key] = value
 
-    def __iter__(self):
-        return iter(self.local_params)
-            
+    def _quote(self, value, replace_words=True):
+        value = unicode(value)
+        if replace_words:
+            value = process_special_words(value)
+        if contains_special_characters(value, self.SPECIAL_CHARACTERS):
+            return "'%s'" % value.replace("'", "\\\\'").replace('"', '\\"')
+        return value
+    
     def __str__(self):
-        if not self.local_params:
+        if not self:
             return ''
 
         parts = []
-        for key, value in self.local_params:
-            if value is None:
-                parts.append(key)
+        for key, value in self.items():
+            if key == 'type':
+                parts.append(value)
             else:
-                parts.append('%s=%s' % (key, value))
+                replace_words = True
+                if isinstance(value, X):
+                    value = make_fq(value)
+                    replace_words = False
+                parts.append(
+                    '%s=%s' % (
+                        key,
+                        self._quote(process_value(value, safe=True),
+                                    replace_words=replace_words)))
         return '{!%s}' % ' '.join(parts)
 
-def process_value(v):
+def process_value(v, safe=False):
     if v is True:
-        return '1'
+        return 'true'
     if v is False:
-        return '0'
+        return 'false'
+    if isinstance(v, LocalParams):
+        return '"%s"' % unicode(v)
     if isinstance(v, (datetime, date)):
         return v.strftime('%Y-%m-%dT%H:%M:%SZ')
     if isinstance(v, basestring) and SOLR_DATETIME_RE.match(v):
         return v
+    if safe:
+        return unicode(v)
     return safe_solr_input(v)
 
 def process_field(field, op, value):
@@ -190,7 +236,9 @@ def make_fq(x, local_params=None):
         for child in x.children:
             if child is None:
                 continue
-            if isinstance(child, tuple):
+            if isinstance(child, LocalParams):
+                parts = [unicode(child)]
+            elif isinstance(child, tuple):
                 parts = [fq_from_tuple(child)]
             elif isinstance(child, basestring):
                 parts = [safe_solr_input(child)]
